
/* ************************************************************************ *
 *              Written by Alex de Kruijff                2009              *
 * ************************************************************************ *
 * This source was written with a tabstop every four characters             *
 * In vi type :set ts=4                                                     *
 * ************************************************************************ */

#include "configure.h"
#include "hash.h"
#include "toolkit.h"
#include "visitor.h"
#include "filename.h"
#include "filegroup.h"

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif // HAVE_FCNTL_H
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif // HAVE_UNISTD_H

#include <new>

size_t tmpCapacity = 256;
#ifdef CHEAP_FILENAME
WrapperFilename FileGroup::tmp(tmpCapacity);
#else // CHEAP_FILENAME
Filename FileGroup::tmp(tmpCapacity);
#endif // CHEAP_FILENAME

void deleteFilename(void *ptr)
{
	delete[] (char *)ptr;
}

#ifndef LIST
hash_t FileGroup::hashFunction(const FileGroup &obj) throw()
{
	return hashword(&obj.st_dev, 1, hashword(&obj.st_ino, 1)); 
}
#endif

int FileGroup::compareNlink(const void *a, const void *b) throw()
{
	FileGroup &fa = **(FileGroup **)a;
	FileGroup &fb = **(FileGroup **)b;
	if (fb.st_nlink - fa.st_nlink)
		return fb.st_nlink - fa.st_nlink;
	if (fa.st_dev - fb.st_dev)
		return fa.st_dev - fb.st_dev;
	return fa.st_ino - fb.st_ino;
}

int FileGroup::compareFirst(const void *a, const void *b) throw()
{
	FileGroup &fa = **(FileGroup **)a;
	FileGroup &fb = **(FileGroup **)b;
#ifdef DEBUG
#ifdef LIST
	if (fa.list.first() == NULL || fb.list.first() == NULL)
#else
	if (fa.hash.first() == NULL || fb.hash.first() == NULL)
#endif
	{
		fprintf(stderr, "%s:%d can not compare NULL\n",
			__FILE__, __LINE__);
		exit(EXIT_FAILURE);
	}
#endif
#ifdef LIST
	return fa.list.first()->compare(*fb.list.first());
#else
	return fa.hash.first()->compare(*fb.hash.first());
#endif
}

int FileGroup::compareLast(const void *a, const void *b) throw()
{
	FileGroup &fa = **(FileGroup **)a;
	FileGroup &fb = **(FileGroup **)b;
#ifdef DEBUG
#ifdef LIST
	if (fa.list.last() == NULL || fb.list.last() == NULL)
#else
	if (fa.hash.last() == NULL || fb.hash.last() == NULL)
#endif
	{
		fprintf(stderr, "%s:%d can not compare NULL\n",
			__FILE__, __LINE__);
		exit(EXIT_FAILURE);
	}
#endif
#ifdef LIST
	return fb.list.last()->compare(*fa.list.last());
#else
	return fb.hash.last()->compare(*fa.hash.last());
#endif
}

int FileGroup::compareOldest(const void *a, const void *b) throw()
{
	FileGroup &fa = **(FileGroup **)a;
	FileGroup &fb = **(FileGroup **)b;
	if (fa.st_mtime - fb.st_mtime)
		return fa.st_mtime - fb.st_mtime;
	if (fa.st_dev - fb.st_dev)
		return fa.st_dev - fb.st_dev;
	return fa.st_ino - fb.st_ino;
}

int FileGroup::compareYoungest(const void *a, const void *b) throw()
{
	FileGroup &fa = **(FileGroup **)a;
	FileGroup &fb = **(FileGroup **)b;
	if (fb.st_mtime - fa.st_mtime)
		return fb.st_mtime - fa.st_mtime;
	if (fa.st_dev - fb.st_dev)
		return fa.st_dev - fb.st_dev;
	return fa.st_ino - fb.st_ino;
}

/* ************************************************************************ */

#ifdef LIST
FileGroup::FileGroup() throw (std::bad_alloc)
#else
FileGroup::FileGroup() throw (std::bad_alloc) : hash(1)
#endif
{
}

FileGroup::FileGroup(const struct stat &s, size_t capacity)
#ifdef LIST
throw (std::bad_alloc)
#else
throw (std::bad_alloc) : hash(capacity)
#endif
{
	st_dev		= s.st_dev;
	st_ino		= s.st_ino;
	st_nlink	= s.st_nlink;
	st_mtime	= s.st_mtime;
#ifndef LIST
	hash.convert(CONTAINER_VECTOR);
#endif
}

void FileGroup::accept(SamefileVisitor &v)
{
	if (v.visit(*this))
		return;
#ifdef LIST
	List<Filename> *l = &list;
	while(l != NULL)
	{
		v.visit(*l->first());
		l = l->next();
	}
#else
	size_t n = hash.getBoundry();
	for (size_t i = 0; i < n; ++i)
		if (hash[i] != NULL)
			v.visit(*hash[i]);
#endif
}


int FileGroup::open(int flags) const throw()
{
#ifdef LIST
	const List<Filename> *l = &list;
	while(l != NULL)
	{
		int fd = ::open(l->first()->data(), flags);
		if (fd >= 0)
			return fd;
		l = l->next();
	}
#else
	size_t n = hash.getBoundry();
	for (size_t i = 0; i < n; ++i)
		if (hash[i] != NULL)
		{
			int fd = ::open(hash[i]->data(), flags);
			if (fd >= 0)
				return fd;
		}
#endif
	return -1;
}

int FileGroup::fcmp(const FileGroup &obj, off_t fileSize) const throw()
{
	// Open the two files
	int fd1 = open(O_RDONLY);
	if (fd1 < 0)
		return FILE_OPEN1_ERROR;

	int fd2 = obj.open(O_RDONLY);
	if (fd2 < 0)
	{
		close(fd1);
		return FILE_OPEN2_ERROR;
	}

	int status = ::fcmp(fd1, fd2, fileSize, fileSize);
	close(fd1);
	close(fd2);
	return status;
}

int FileGroup::operator!=(const char *path) const throw (std::bad_alloc)
{
	size_t len = strlen(path);
	if (tmpCapacity < len)
	{
		while(tmpCapacity < len)
			tmpCapacity <<= 1;
		tmp.renew(tmpCapacity);
	}
	tmp = path;
#ifdef LIST
	return list != tmp.getFilename();
#else
	return hash != tmp;
#endif
}

void FileGroup::operator+=(const char *path) throw (std::bad_alloc)
{
#ifdef CHEAP_FILENAME
	size_t len = strlen(path);
	if (!len)
		return;
	char *ptr = new char[++len];
	memcpy(ptr, path, len);
	try
	{
#ifdef LIST
		list += (Filename *)ptr; // throws bad_alloc
#else
		hash += *(Filename *)ptr; // throws bad_alloc
#endif
	}
#else // CHEAP_FILENAME
	Filename *ptr = new Filename(path); // throws bad_alloc
	try
	{
#ifdef LIST
		list += ptr; // throws bad_alloc
#else
		hash += *ptr; // throws bad_alloc
#endif
	}
#endif // CHEAP_FILENAME
	catch(std::bad_alloc &e)
	{
		delete ptr;
		throw(e);
	}
}

