
/* ************************************************************************ *
 *              Written by Alex de Kruijff                2009              *
 * ************************************************************************ *
 * This source was written with a tabstop every four characters             *
 * In vi type :set ts=4                                                     *
 * ************************************************************************ */

#ifndef AK_FILEGROUP_H
#define AK_FILEGROUP_H

#include "configure.h"
#include "hash.h"
#include "container.h"
#ifdef LIST
#include "list.h"
#endif
// #include "storage.h"
// #include "visitor.h"
#include "filename.h"

#ifdef DEBUG
#include <stdio.h>
#endif

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif // HAVE_SYS_STAT_H

#include <new>

class SizeGroup;
class SamefileVisitor;

#ifdef CHEAP_FILENAME
void deleteFilename(void *ptr);
#endif // CHEAP_FILENAME

/**
 * A FileGroup is a group of Filename(s) that share the same device and inode
 * numers.
 * @see Filename, Container
 */
class FileGroup
{
#ifdef CHEAP_FILENAME
	static WrapperFilename tmp;
#else // CHEAP_FILENAME
	static Filename tmp;
#endif // CHEAP_FILENAME

public:
	__uint32_t	st_dev;
	ino_t		st_ino;
	nlink_t		st_nlink;
#if __BSD_VISIBLE
	struct  timespec st_mtimespec;
#else 
	time_t		st_mtime;
#endif

private:
#ifndef LIST
	Container<Filename> hash;
#else
	List<Filename> list;
#endif

	friend class SizeGroup;

public:
#ifndef LIST
	static hash_t hashFunction(const FileGroup &obj) throw();
#endif

	/**
	 * Assumes a and b are of type FileGroup ** and compares them on the
	 * number of links. If those are equal then the first Filename is 
	 * used for comparison.
	 */
	static int compareNlink(const void *a, const void *b) throw();

	/**
	 * Assumes a and b are of type FileGroup ** and compares them 
	 * alphabetically using the first Filename.
	 * In order for this to work the function sort must be called.
	 * @see sort
	 */
	static int compareFirst(const void *a, const void *b) throw();

	/**
	 * Assumes a and b are of type FileGroup ** and compares them 
	 * reversed alphabetically using the last Filename. 
	 * The function sort must be called in order for this to work.
	 * @see sort
	 */
	static int compareLast(const void *a, const void *b) throw();

	/**
	 * Assumes a and b are of type FileGroup ** and compares them 
	 * chronologically using the modication time.
	 * The function sort must be called in order for this to work.
	 * @see sort
	 */
	static int compareOldest(const void *a, const void *b) throw();

	/**
	 * Assumes a and b are of type FileGroup ** and compares them 
	 * reversed chronologically using the modication time.
	 * The function sort must be called in order for this to work.
	 * @see sort
	 */
	static int compareYoungest(const void *a, const void *b) throw();

	/**
	 * Creates a FileGroup object without a preset shared metadata.
	 */
	FileGroup() throw (std::bad_alloc);

	/**
	 * Creates a FileGroup object with a preset shared meta data.
	 * @param s - shared meta data 
	 * @param capacity - the initial capacity of the group.
	 */
	FileGroup(const struct stat &s, size_t capacity = 1)
	throw (std::bad_alloc);

#ifndef LIST
#ifdef CHEAP_FILENAME
	~FileGroup() throw() { hash.deleteItems(&deleteFilename); }
#else // CHEAP_FILENAME
	~FileGroup() throw() { hash.deleteItems(); }
#endif // CHEAP_FILENAME
#else
#ifdef CHEAP_FILENAME
	~FileGroup() throw() { list.deleteItems(&deleteFilename); }
#else // CHEAP_FILENAME
	~FileGroup() throw() { list.deleteItems(); }
#endif // CHEAP_FILENAME
#endif

	/**
	 *
	 */
#ifndef LIST
	void empty() throw() { hash.empty(1); }
#else
	void empty() throw() { list.empty(1); }
#endif

#ifdef DEBUG
#ifndef LIST
	size_t getCapacity() const throw() { return hash.getCapacity(); }
#else
	size_t getCapacity() const throw() { return 0; }
#endif
#endif
	/**
	 * Return the number of Filenames within this group.
	 */
#ifndef LIST
	size_t getSize() const throw() { return hash.getSize(); }
	int isEmpty() const throw() { return !hash.getSize(); }
#else
	size_t getSize() const throw() { return list.getSize(); }
	int isEmpty() const throw() { return list.isEmpty(); }
#endif

#ifndef LIST
        /**
         * Gets the boundry for the mode of the container.
         */   
        size_t getBoundry() const throw() { return hash.getBoundry(); }
#else
        size_t getBoundry() const throw() { return list.getSize(); }
#endif

#ifdef LIST
	ListIterator<Filename> *createIterator() const
	{ return new ListIterator<Filename>(list);}
#endif

	/**
	 * Implements the visitor pattern.
	 */
	void accept(SamefileVisitor &v);

	/**
	 * Checks if both files are on the same device.
	 */
	int isOnSameDevice(const FileGroup &obj) const throw()
	{ return st_dev == obj.st_dev; }

	/**
	 * Sort all the Filename on alphabethical order.
	 */
	void sort() throw()
#ifdef LIST
	{ list.sort(Filename::compare); };
#else
	{ hash.sort(Filename::compare); };
#endif

	/**
	 * Open the file using any of Filename(s) within this group.
	 */
	int open(int flags) const throw();

	/**
	 * Compares if this FileGroup has the same contence of the given
	 * FileGroup.
	 * @returns see ::fcmp()
	 */
	int fcmp(const FileGroup &obj, off_t fileSize) const throw();

	/**
	 * Finds the Filename at the given index.
	 * @return NULL if the above is not true.
	 */
#ifndef LIST
	const Filename *operator[](int index) const
	{ return hash[index]; }
#endif

	/**
	 * Is this the same dev, inode pair?
	 */
	int operator==(const FileGroup &obj) const throw()
	{ return st_ino == obj.st_ino && st_dev == obj.st_dev; }

	/**
	 * Is this the same dev, inode pair?
	 */
	int operator!=(const FileGroup &obj) const throw()
	{ return st_ino != obj.st_ino || st_dev != obj.st_dev; }

	/**
	 * Doesn't str live in this object?
	 */
	int operator!=(const char *str) const throw (std::bad_alloc);

	/**
	 * Does str live in this object?
	 */
	int operator==(const char *str) const throw (std::bad_alloc)
	{ return !operator!=(str); }

	/**
	 * Sets the meta data of this FileGroup to s.
	 */
	void operator=(const struct stat &s) throw()
	{ st_ino = s.st_ino, st_dev = s.st_dev; }

	/**
	 * Adds a Filename to the FileGroup.
	 */
	void operator+=(const char *path) throw (std::bad_alloc);
};

#endif // AK_FILEGROUP_H

